@@ -122,7 +122,7 @@ vec_default_cast <- function(x, to, ..., x_arg = "", to_arg = "") {
 
   opts <- match_fallback_opts(...)
 
-  if (opts$s3_fallback && is_common_class_fallback(to) && length(common_class_suffix(x, to))) {
+  if (is_common_class_fallback(to) && length(common_class_suffix(x, to))) {
     return(x)
   }
 
